﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace InteractiveJewels
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        Jewel[,] jewels = new Jewel[8, 8];
        string[] stacks = new string[8];
        int[] currpos = new int[8];
        bool inGame = false;
        bool animating = false;

        Jewel[] nextUpJewels = new Jewel[8];

        public MainWindow()
        {
            InitializeComponent();

            for (int i = 0; i < 8; i++)
            {
                for (int j = 0; j < 8; j++)
                {
                    Jewel jew = new Jewel();
                    Canvas.SetLeft(jew, 500 * i);
                    Canvas.SetTop(jew, 500 * j);
                    jew.X = i;
                    jew.Y = j;
                    MainCanvas.Children.Add(jew);
                    jewels[i, j] = jew;

                }
                Jewel next = new Jewel();
                Canvas.SetLeft(next, 500 * i);
                NextUpCanvas.Children.Add(next);
                next.Opacity = 1;
                nextUpJewels[i] = next;
            }
        }

        private void Button_Click(object sender, RoutedEventArgs e)
        {
            NewGame ng = new NewGame();
            if (ng.ShowDialog() == true)
            {
                StreamReader sr = new StreamReader(ng.fileName.Text);
                for (int j = 0; j < 8; j++)
                {
                    string input = sr.ReadLine();
                    for (int i = 0; i < 8; i++)
                    {
                        jewels[i, j].JewelColor = input[i] - '0';
                    }
                }
                for (int i = 0; i < 8; i++)
                {
                    stacks[i] = sr.ReadLine();
                    nextUpJewels[i].JewelColor = stacks[i][0] - '0';
                }
                sr.Close();
                totalScore = 0;
                txtScore.Text = "Skóre: 0";
                inGame = true;
                GameOver.Visibility = Visibility.Hidden;

                var moves = getPossibleMoves();
                if (moves.Count == 0)
                {
                    GameOver.Visibility = Visibility.Visible;
                    inGame = false;
                }
                else
                {
                    drawHints(moves);
                }
            }
        }

        private Jewel selectedJewel = null;
        private int totalScore = 0;

        private async void MainCanvas_PreviewMouseUp(object sender, MouseButtonEventArgs e)
        {
            if (inGame && e.Source is Jewel && !animating)
            {
                Jewel jew = (Jewel)e.Source;
                if (!jew.IsReachable) return;
                if (selectedJewel == null)
                {
                    jew.IsSelected = true;
                    selectedJewel = jew;
                    for (int i = 0; i < 8; i++)
                    {
                        for (int j = 0; j < 8; j++)
                        {
                            jewels[i, j].IsReachable = false;
                        }
                    }

                    jew.IsReachable = true;
                    if (jew.X > 0) jewels[jew.X - 1, jew.Y].IsReachable = true;
                    if (jew.X < 7) jewels[jew.X + 1, jew.Y].IsReachable = true;
                    if (jew.Y > 0) jewels[jew.X, jew.Y - 1].IsReachable = true;
                    if (jew.Y < 7) jewels[jew.X, jew.Y + 1].IsReachable = true;
                }
                else
                {
                    if (selectedJewel == jew)
                    {
                        for (int i = 0; i < 8; i++)
                        {
                            for (int j = 0; j < 8; j++)
                            {
                                jewels[i, j].IsReachable = true;
                            }
                        }
                        selectedJewel = null;
                        jew.IsSelected = false;
                    }
                    else
                    {
                        animating = true;
                        for (int i = 0; i < 8; i++)
                        {
                            nextUpJewels[i].Opacity = 0.4;
                        }
                        HintsCanvas.Children.Clear();
                        for (int i = 0; i < 8; i++)
                        {
                            for (int j = 0; j < 8; j++)
                            {
                                jewels[i, j].IsReachable = true;
                            }
                        }
                        selectedJewel.IsSelected = false;
                        await swap(jew, selectedJewel);
                        int score = processTriples();
                        if (score == 0)
                        {
                            await swap(jew, selectedJewel);
                        }
                        else
                        {
                            totalScore += score;
                            txtScore.Text = "Skóre: " + totalScore;
                            await remove();
                            while ((score = processTriples()) > 0)
                            {
                                totalScore += score;
                                txtScore.Text = "Skóre: " + totalScore;
                                await remove();
                            }
                        }

                        var moves = getPossibleMoves();
                        if (moves.Count == 0)
                        {
                            GameOver.Visibility = Visibility.Visible;
                            inGame = false;
                        }
                        else
                        {
                            drawHints(moves);
                        }

                        for(int i = 0;i<8;i++)
                        {
                            nextUpJewels[i].JewelColor = stacks[i][currpos[i]] - '0';
                            nextUpJewels[i].Opacity = 1;
                        }

                        selectedJewel = null;
                        animating = false;

                    }

                }
            }
        }

        private Tuple<Point, Point> pointTuple(int x1, int y1, int x2, int y2)
        {
            return new Tuple<Point, Point>(new Point(x1, y1), new Point(x2, y2));
        }

        private List<Tuple<Point, Point>> getPossibleMoves()
        {
            List<Tuple<Point, Point>> results = new List<Tuple<Point, Point>>();

            //Po řádcích
            for (int j = 0; j < 8; j++)
            {
                int len = 1;
                int currentColor = jewels[0, j].JewelColor;
                for (int i = 1; i < 8; i++)
                {
                    if (jewels[i, j].JewelColor == currentColor)
                    {
                        len++;
                    }
                    else
                    {
                        if (len >= 2)
                        {
                            if (i-1 >= 3 && jewels[i-1 - 3, j].JewelColor == currentColor) results.Add(pointTuple(i-1 - 3, j, i-1 - 2, j));
                            if (i-1 >= 2)
                            {
                                if (j >= 1 && jewels[i-1 - 2, j - 1].JewelColor == currentColor) results.Add(pointTuple(i-1 - 2, j - 1, i-1 - 2, j));
                                if (j <= 6 && jewels[i-1 - 2, j + 1].JewelColor == currentColor) results.Add(pointTuple(i-1 - 2, j + 1, i-1 - 2, j));
                            }
                            if (i-1 <= 6)
                            {
                                if (j >= 1 && jewels[i-1 + 1, j - 1].JewelColor == currentColor) results.Add(pointTuple(i-1 + 1, j, i-1 + 1, j - 1));
                                if (j <= 6 && jewels[i-1 + 1, j + 1].JewelColor == currentColor) results.Add(pointTuple(i-1 + 1, j, i-1 + 1, j + 1));
                            }
                            if (i-1 <= 5 && jewels[i-1 + 2, j].JewelColor == currentColor) results.Add(pointTuple(i-1 + 1, j, i-1 + 2, j));
                        }
                        len = 1;
                        currentColor = jewels[i, j].JewelColor;
                    }

                    if (i >= 1 && i <= 6 && jewels[i - 1, j].JewelColor == jewels[i + 1, j].JewelColor)
                    {
                        if (j >= 1 && jewels[i, j - 1].JewelColor == jewels[i - 1, j].JewelColor) results.Add(pointTuple(i, j - 1, i, j));
                        if (j <= 6 && jewels[i, j + 1].JewelColor == jewels[i - 1, j].JewelColor) results.Add(pointTuple(i, j + 1, i, j));

                    }

                }
            }


            //Po Sloupcích
            for (int i = 0; i < 8; i++)
            {
                int len = 1;
                int currentColor = jewels[i, 0].JewelColor;
                for (int j = 1; j < 8; j++)
                {
                    if (jewels[i, j].JewelColor == currentColor)
                    {
                        len++;
                    }
                    else
                    {
                        if (len >= 2)
                        {
                            if (j-1 >= 3 && jewels[i, j-1 - 3].JewelColor == currentColor) results.Add(pointTuple(i, j-1 - 3, i, j-1 - 2));
                            if (j-1 >= 2)
                            {
                                if (i >= 1 && jewels[i - 1, j-1 - 2].JewelColor == currentColor) results.Add(pointTuple(i - 1, j-1 - 2, i, j-1 - 2));
                                if (i <= 6 && jewels[i + 1, j-1 - 2].JewelColor == currentColor) results.Add(pointTuple(i + 1, j-1 - 2, i, j-1 - 2));
                            }
                            if (j-1 <= 6)
                            {
                                if (i >= 1 && jewels[i - 1, j-1 + 1].JewelColor == currentColor) results.Add(pointTuple(i, j-1 + 1, i - 1, j-1 + 1));
                                if (i <= 6 && jewels[i + 1, j-1 + 1].JewelColor == currentColor) results.Add(pointTuple(i, j-1 + 1, i + 1, j-1 + 1));
                            }
                            if (j-1 <= 5 && jewels[i, j-1 + 2].JewelColor == currentColor) results.Add(pointTuple(i, j-1 + 1, i, j-1 + 2));
                        }
                        len = 1;
                        currentColor = jewels[i, j].JewelColor;
                    }
                    if (j >= 1 && j <= 6 && jewels[i, j - 1].JewelColor == jewels[i, j + 1].JewelColor)
                    {
                        if (i >= 1 && jewels[i - 1, j].JewelColor == jewels[i, j - 1].JewelColor) results.Add(pointTuple(i - 1, j, i, j));
                        if (i <= 6 && jewels[i + 1, j].JewelColor == jewels[i, j - 1].JewelColor) results.Add(pointTuple(i + 1, j, i, j));

                    }
                }
            }



            return results;
        }

        private async Task swap(Jewel jew1, Jewel jew2)
        {
            TimeSpan duration = TimeSpan.FromMilliseconds(300);
            Storyboard sb = new Storyboard();
            sb.Duration = duration;
            sb.FillBehavior = FillBehavior.Stop;

            DoubleAnimation anim1 = new DoubleAnimation();
            anim1.To = Canvas.GetLeft(jew1);
            anim1.Duration = duration;
            Storyboard.SetTargetProperty(anim1, new PropertyPath("(Canvas.Left)"));
            Storyboard.SetTarget(anim1, jew2);

            DoubleAnimation anim2 = new DoubleAnimation();
            anim2.To = Canvas.GetTop(jew1);
            anim2.Duration = duration;
            Storyboard.SetTargetProperty(anim2, new PropertyPath("(Canvas.Top)"));
            Storyboard.SetTarget(anim2, jew2);

            DoubleAnimation anim3 = new DoubleAnimation();
            anim3.To = Canvas.GetLeft(jew2);
            anim3.Duration = duration;
            Storyboard.SetTargetProperty(anim3, new PropertyPath("(Canvas.Left)"));
            Storyboard.SetTarget(anim3, jew1);

            DoubleAnimation anim4 = new DoubleAnimation();
            anim4.To = Canvas.GetTop(jew2);
            anim4.Duration = duration;
            Storyboard.SetTargetProperty(anim4, new PropertyPath("(Canvas.Top)"));
            Storyboard.SetTarget(anim4, jew1);

            sb.Children.Add(anim1);
            sb.Children.Add(anim2);
            sb.Children.Add(anim3);
            sb.Children.Add(anim4);

            await StoryBoardAsync(sb);

            int tmp = jew1.JewelColor;
            jew1.JewelColor = jew2.JewelColor;
            jew2.JewelColor = tmp;



        }

        private void drawHints(List<Tuple<Point, Point>> hints)
        {
            HintsCanvas.Children.Clear();

            foreach (Tuple<Point, Point> hint in hints)
            {
                UserControl us;
                if (hint.Item1.X != hint.Item2.X)
                {
                    us = new HorizontalSwap();
                    Canvas.SetLeft(us, Math.Max(hint.Item1.X, hint.Item2.X) * 500 - 75);
                    Canvas.SetTop(us, Math.Max(hint.Item1.Y, hint.Item2.Y) * 500 + 175);
                }
                else
                {
                    us = new VerticalSwap();
                    Canvas.SetLeft(us, Math.Max(hint.Item1.X, hint.Item2.X) * 500 + 175);
                    Canvas.SetTop(us, Math.Max(hint.Item1.Y, hint.Item2.Y) * 500 - 75);
                }

                HintsCanvas.Children.Add(us);
            }

        }

        async private Task remove()
        {
            Storyboard disappearsb = new Storyboard();
            disappearsb.FillBehavior = FillBehavior.Stop;


            for (int i = 0; i < 8; i++)
            {
                for (int j = 7; j >= 0; j--)
                {
                    if (jewels[i, j].IsRemoved)
                    {
                        DoubleAnimation anim = new DoubleAnimation();
                        anim.To = 0;
                        anim.Duration = TimeSpan.FromMilliseconds(200);
                        Storyboard.SetTargetProperty(anim, new PropertyPath("Opacity"));
                        Storyboard.SetTarget(anim, jewels[i, j]);
                        disappearsb.Children.Add(anim);
                    }
                }
            }

            await StoryBoardAsync(disappearsb);

            int duration = 200;
            int maxfall = 0;
            Storyboard fallsb = new Storyboard();
            disappearsb.FillBehavior = FillBehavior.Stop;
            for (int i = 0; i < 8; i++)
            {
                int fallOffset = 0;
                for (int j = 7; j >= 0; j--)
                {
                    if (jewels[i, j].IsRemoved)
                    {
                        for (int x = j - 1; x >= -1; x--)
                        {
                            if (x == -1)
                            {
                                jewels[i, j].JewelColor = stacks[i][currpos[i]] - '0';
                                currpos[i]++;
                                if (currpos[i] >= stacks[i].Length) currpos[i] = 0;
                                fallOffset++;

                                int fallHeight = jewels[i, j].Y + fallOffset;
                                if (fallHeight > maxfall) maxfall = fallHeight;
                                DoubleAnimation anim = new DoubleAnimation();
                                anim.From = (-500) * fallOffset;
                                anim.To = Canvas.GetTop(jewels[i, j]);

                                anim.Duration = TimeSpan.FromMilliseconds(fallHeight * duration);
                                Storyboard.SetTargetProperty(anim, new PropertyPath("(Canvas.Top)"));
                                Storyboard.SetTarget(anim, jewels[i, j]);
                                fallsb.Children.Add(anim);
                                break;
                            }
                            else if (!jewels[i, x].IsRemoved)
                            {
                                jewels[i, j].JewelColor = jewels[i, x].JewelColor;
                                int fallHeight = jewels[i, j].Y - jewels[i, x].Y;
                                if (fallHeight > maxfall) maxfall = fallHeight;

                                DoubleAnimation anim = new DoubleAnimation();
                                anim.From = Canvas.GetTop(jewels[i, x]);
                                anim.To = Canvas.GetTop(jewels[i, j]);
                                anim.Duration = TimeSpan.FromMilliseconds(fallHeight * duration);
                                Storyboard.SetTargetProperty(anim, new PropertyPath("(Canvas.Top)"));
                                Storyboard.SetTarget(anim, jewels[i, j]);
                                fallsb.Children.Add(anim);
                                jewels[i, x].IsRemoved = true;
                                break;
                            }
                        }
                    }
                }
            }

            fallsb.Duration = TimeSpan.FromMilliseconds(duration * maxfall);
            await StoryBoardAsync(fallsb);

            for (int i = 0; i < 8; i++)
            {
                for (int j = 0; j < 8; j++)
                {
                    jewels[i, j].IsRemoved = false;
                }
            }
        }

        private int processTriples()
        {

            int score = 0;

            //Po řádcích
            for (int j = 0; j < 8; j++)
            {
                int len = 1;
                int currentColor = jewels[0, j].JewelColor;
                for (int i = 1; i < 8; i++)
                {
                    if (jewels[i, j].JewelColor == currentColor)
                    {
                        len++;
                    }
                    else
                    {
                        len = 1;
                        currentColor = jewels[i, j].JewelColor;
                    }
                    if (len >= 3)
                    {
                        score += 3;
                        jewels[i, j].IsRemoved = true;
                        jewels[i - 1, j].IsRemoved = true;
                        jewels[i - 2, j].IsRemoved = true;
                    }
                }
            }

            //Po Sloupcích
            for (int i = 0; i < 8; i++)
            {
                int len = 1;
                int currentColor = jewels[i, 0].JewelColor;
                for (int j = 1; j < 8; j++)
                {
                    if (jewels[i, j].JewelColor == currentColor)
                    {
                        len++;
                    }
                    else
                    {
                        len = 1;
                        currentColor = jewels[i, j].JewelColor;
                    }
                    if (len >= 3)
                    {
                        score += 3;
                        jewels[i, j].IsRemoved = true;
                        jewels[i, j - 1].IsRemoved = true;
                        jewels[i, j - 2].IsRemoved = true;
                    }
                }
            }

            return score;
        }

        public static Task StoryBoardAsync(Storyboard storyboard)
        {
            System.Threading.Tasks.TaskCompletionSource<bool> tcs = new TaskCompletionSource<bool>();
            if (storyboard == null)
                tcs.SetException(new ArgumentNullException());
            else
            {
                EventHandler onComplete = null;
                onComplete = (s, e) =>
                {
                    storyboard.Completed -= onComplete;
                    tcs.SetResult(true);
                };
                storyboard.Completed += onComplete;
                storyboard.Begin();
            }
            return tcs.Task;
        }

        private void CheckBox_Checked(object sender, RoutedEventArgs e)
        {
            HintsCanvas.Visibility = Visibility.Visible;
        }

        private void CheckBox_Unchecked(object sender, RoutedEventArgs e)
        {
            HintsCanvas.Visibility = Visibility.Hidden;
        }
    }


}
